/*
 * Copyright (c) 2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Sep  8 2021
 */
#include "acxsensmvn.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <poll.h>
#include <regex.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include "codels.h"


static int		mvn_remote_cmd(int fd, const char *req, const char *ack,
                                struct timeval deadline, size_t length,
                                char reply[], size_t nmatch,
                                regmatch_t pmatch[]);

const char *mvnsegname[] = {
  [ 0] = "pelvis",
  [ 1] = "l5",
  [ 2] = "l3",
  [ 3] = "t12",
  [ 4] = "t8",
  [ 5] = "neck",
  [ 6] = "head",
  [ 7] = "rshoulder",
  [ 8] = "rarm",
  [ 9] = "rforearm",
  [10] = "rand",
  [11] = "lshoulder",
  [12] = "larm",
  [13] = "lforearm",
  [14] = "lhand",
  [15] = "rupleg",
  [16] = "rloleg",
  [17] = "rfoot",
  [18] = "rtoe",
  [19] = "lupleg",
  [20] = "lloleg",
  [21] = "lfoot",
  [22] = "ltoe",
  [24] = "prop1",
  [25] = "prop2",
  [26] = "prop3",
  [27] = "prop4"
};

size_t const mvnsegnames = sizeof(mvnsegname)/sizeof(*mvnsegname);


/* --- mvn_remote_socket --------------------------------------------------- */

/** Create a socket talking to the MVN 'remote control' port
 */
int
mvn_remote_socket(const char *host, const char *port, struct timeval deadline,
                  char *emsg, size_t emsgsize)
{
  struct addrinfo hints, *res, *res0;
  char loghost[64], logport[16];
  const char *cause = "error";
  int fd;
  int e;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET; /* MVN does not support IPv6 */
  hints.ai_socktype = SOCK_DGRAM;
  e = getaddrinfo(host, port, &hints, &res0);
  if (e) {
    snprintf(emsg, emsgsize, "%s", gai_strerror(e));
    return -1;
  }

  for (res = res0; res; res = res->ai_next) {
    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd < 0) {
      cause = "socket"; goto log;
    }

    /* connect */
    if (connect(fd, res->ai_addr, res->ai_addrlen)) {
      cause = "connect"; goto log;
    }

    /* send nothing to detect local errors */
    if (send(fd, NULL, 0, 0)) {
      cause = "send"; goto log;
    }

    /* ping */
    if (mvn_remote_ping(fd, deadline) < 0) {
      cause = "ping"; goto log;
    }

    /* success */
    break;

log:
    e = errno;
    if (fd >= 0) close(fd);
    fd = -1;

    if (!getnameinfo(res->ai_addr, res->ai_addrlen,
                     loghost, sizeof(loghost),
                     logport, sizeof(logport), NI_NUMERICHOST|NI_NUMERICSERV))
      snprintf(emsg, emsgsize,
               "%s on %s.%s: %s", cause, loghost, logport, strerror(e));
  }

  if (res0) freeaddrinfo(res0);
  return fd;
}


/* --- mvn_remote_localhost ------------------------------------------------ */

/** Return a socket address suitable to send stream data to
 */
int
mvn_remote_localhost(int fd, char *host, size_t hostlen,
                     char *emsg, size_t emsgsize)
{
  struct sockaddr_storage addr;
  socklen_t len = sizeof(addr);
  const char *cause = "error";

  if (getsockname(fd, (struct sockaddr *)&addr, &len)) {
    cause = "getsockname"; goto err;
  }

  if (getnameinfo((struct sockaddr *)&addr, len,
                  host, hostlen, NULL, 0, NI_NUMERICHOST)) {
    cause = "getnameinfo"; goto err;
  }

  return 0;

err:
  snprintf(emsg, emsgsize, "%s: %s", cause, strerror(errno));
  return -1;
}


/* --- mvn_remote_ping ----------------------------------------------------- */

/** Send IdentifyReq and wait for reply
 */
int
mvn_remote_ping(int fd, struct timeval deadline)
{
  static const char req[] = "<IdentifyReq/>";
  static const char ack[] = "<IdentifyAck[^>]*InstanceName=\"([^\"]*)\"";

  char reply[256];
  regmatch_t pmatch[2];

  if (mvn_remote_cmd(fd, req, ack, deadline, sizeof(reply), reply, 2, pmatch))
    return -1;

  if (pmatch[1].rm_so != -1 && pmatch[1].rm_eo > pmatch[1].rm_so) {
    reply[pmatch[1].rm_eo] = 0;
    warnx("detected remote instance '%s'", &reply[pmatch[1].rm_so]);
  }

  return 0;
}


/* --- mvn_remote_stream --------------------------------------------------- */

/** Send AddNetworkStreamingTargetReq and wait for reply
 */
int
mvn_remote_stream(int fd, bool add,
                  const char *host, const char *port, const char *proto,
                  struct timeval deadline)
{
  static const char req[] =
    "<%sNetworkStreamingTargetReq "
    "IpAddress=\"%s\" PortNumber=\"%s\" Protocol=\"%s\"/>";
  static const char ack[] =
    "<(Add|Remove)NetworkStreamingTargetAck[^>]*Result=\"([^\"]*)\"";

  char cmd[256], reply[128];
  regmatch_t pmatch[3];
  int s;

  s = snprintf(
    cmd, sizeof(cmd), req, add ? "Add" : "Remove", host, port, proto);
  if (s < 0) return -1;
  assert((unsigned)s <= sizeof(cmd));

  if (mvn_remote_cmd(fd, cmd, ack, deadline, sizeof(reply), reply, 3, pmatch))
    return -1;

  if (pmatch[2].rm_so == -1) return -1;
  reply[pmatch[2].rm_eo] = 0;

  return strcasecmp(&reply[pmatch[2].rm_so], "true") ? -1 : 0;
}


/* --- mvn_remote_cmd ------------------------------------------------------ */

/** Send remote command and wait for reply
 */
static int
mvn_remote_cmd(int fd, const char *req, const char *ack,
               struct timeval deadline, size_t length, char reply[],
               size_t nmatch, regmatch_t pmatch[])
{
  struct timeval n;
  regex_t preg;
  int timeout;
  ssize_t s;

  if (regcomp(&preg, ack, REG_EXTENDED | REG_ICASE)) {
    warnx("internal regexp error");
    errno = EINTR;
    return -1;
  }

  errno = ETIME;
  for(;;) {
    /* send request */
    if (send(fd, req, strlen(req), 0) < 0)
      break;

    /* poll reply */
    gettimeofday(&n, NULL);
    timeout =
      (deadline.tv_sec - n.tv_sec) * 1000
      + (deadline.tv_usec - n.tv_usec) / 1000;
    if (timeout < 0 ||
        poll(&(struct pollfd){ .fd = fd, .events = POLLIN }, 1, timeout) < 1)
      break;

    s = recv(fd, reply, length, 0);
    if (s < 0) break;

    /* parse reply */
    reply[s] = 0;
    if (regexec(&preg, reply, nmatch, pmatch, 0) == REG_NOMATCH) {
      errno = ENOMSG;
      continue;
    }

    regfree(&preg);
    return 0; /* success */
  }

  regfree(&preg);
  return -1;
}


/* --- mvn_stream_socket --------------------------------------------------- */

/** Create a socket receiving MXTP data
 */
int
mvn_stream_socket(const char *port, char *emsg, size_t emsgsize)
{
  struct addrinfo hints, *res, *res0;
  char loghost[64], logport[16];
  const char *cause;
  int fd;
  int e;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET; /* MVN does not support IPv6 */
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_flags = AI_PASSIVE;
  e = getaddrinfo(NULL, port, &hints, &res0);
  if (e) {
    snprintf(emsg, emsgsize, "%s", gai_strerror(e));
    return -1;
  }

  for (res = res0; res; res = res->ai_next) {
    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd < 0) {
      cause = "socket"; goto log;
    }

    /* try binding on local port */
    if (bind(fd, res->ai_addr, res->ai_addrlen)) {
      cause = "bind"; goto log;
    }

    /* success */
    break;

log:
    e = errno;
    if (fd > 0) close(fd);
    fd = -1;

    if (!getnameinfo(res->ai_addr, res->ai_addrlen,
                     loghost, sizeof(loghost),
                     logport, sizeof(logport), NI_NUMERICHOST|NI_NUMERICSERV))
      snprintf(emsg, emsgsize,
               "%s on %s.%s: %s", cause, loghost, logport, strerror(e));
  }

  /* error */
  if (res0) freeaddrinfo(res0);
  return fd;
}


/* --- mvn_stream_gethost -------------------------------------------------- */

/** Return a socket address suitable to send stream data to
 */
int
mvn_stream_gethost(char *host, size_t hostlen,
                   char *port, size_t portlen,
                   char *emsg, size_t emsgsize)
{
  struct addrinfo hints = {}, *res;
  int e;

  /* empty host is treated specially and kept empty */

  hints.ai_family = AF_INET; /* MVN does not support IPv6 */
  hints.ai_socktype = SOCK_DGRAM;
  e = getaddrinfo(host[0] ? host : NULL, port, &hints, &res);
  if (e) {
    snprintf(emsg, emsgsize, "getaddrinfo: %s", gai_strerror(e));
    return -1;
  }

  e = getnameinfo(
    res->ai_addr, res->ai_addrlen,
    host, host[0] ? hostlen : 0,
    port, portlen, NI_NUMERICHOST | NI_NUMERICSERV);

  freeaddrinfo(res);
  if (e) {
    snprintf(emsg, emsgsize, "getnameinfo: %s", gai_strerror(e));
    return -1;
  }
  return 0;
}


/* --- mvn_stream_decode --------------------------------------------------- */

/** Decode data
 */

struct __attribute__((packed)) mxtp_poseq {
    uint32_t id;
    float x, y, z;
    float qw, qx, qy, qz;
};

struct __attribute__((packed)) mxtp_linseg {
    uint32_t id;
    float x, y, z;
    float vx, vy, vz;
    float ax, ay, az;
};

struct __attribute__((packed)) mxtp_angseg {
    uint32_t id;
    float qw, qx, qy, qz;
    float wx, wy, wz;
    float awx, awy, awz;
};

struct __attribute__((packed)) mxtp_msg {
  struct __attribute__((packed)) {
    char sync[4];
    uint8_t type[2];
    uint32_t pkts;
    uint8_t sub;
    uint8_t nitems;
    uint32_t tcode;
    uint8_t id;
    uint8_t nsegs, nprops, nfingers;
    uint8_t pad[2];
    uint16_t sz;
    uint8_t payload[];
  };

  union {
    struct mxtp_poseq poseq[32];	/* 32 is arbitrary, actually nitems */
    struct mxtp_linseg linseg[32];
    struct mxtp_angseg angseg[32];
  };
};

static inline int
mvn_typecode(const uint8_t x[2])
{
  return x[0]*10 + x[1] - '0'*11;
}

static inline float
mvn_ntohf(float x)
{
  union { float f; uint32_t d; } y = { .f = x };
  y.d = ntohl(y.d);
  return y.f;
}


int
mvn_stream_decode(const void *data, size_t len, xsensmvn_body_s *bodies)
{
  const struct mxtp_msg *msg = data;
  struct timeval tv;
  uint32_t seq, id;
  int i;

  /* check message contents */
  if (strncmp(msg->sync, "MXTP", 4)) return -1;
  if (offsetof(struct mxtp_msg, payload[ntohs(msg->sz)]) != len) return -1;

  /* update timestamps */
  seq = ntohl(msg->pkts);
  if (bodies->seq != seq) {
    gettimeofday(&tv, NULL);
    bodies->ts = (struct or_time_ts){
      .sec = tv.tv_sec, .nsec = tv.tv_usec * 1000
    };
    bodies->seq = seq;
    bodies->lin = bodies->ang = false;
    for(id = 0; id < bodies->segments._length; id++)
      bodies->segments._buffer[id] = (xsensmvn_segment_s){
        .pos._present = false,
        .att._present = false,
        .vel._present = false,
        .avel._present = false,
        .acc._present = false,
        .aacc._present = false
      };
  }

  /* decode */
  switch(mvn_typecode(msg->type)) {
    case 21: /* Linear segment kinematics */
      for(i = 0; i < msg->nitems; i++) {
        id = ntohl(msg->linseg[i].id) - 1;
        if (id >= bodies->segments._length) continue;

        bodies->segments._buffer[id].pos = (optional_or_t3d_pos){
          ._present = true,
          ._value.x = mvn_ntohf(msg->linseg[i].x),
          ._value.y = mvn_ntohf(msg->linseg[i].y),
          ._value.z = mvn_ntohf(msg->linseg[i].z)
        };
        bodies->segments._buffer[id].vel = (optional_or_t3d_vel){
          ._present = true,
          ._value.vx = mvn_ntohf(msg->linseg[i].vx),
          ._value.vy = mvn_ntohf(msg->linseg[i].vy),
          ._value.vz = mvn_ntohf(msg->linseg[i].vz)
        };
        bodies->segments._buffer[id].acc = (optional_or_t3d_acc){
          ._present = true,
          ._value.ax = mvn_ntohf(msg->linseg[i].ax),
          ._value.ay = mvn_ntohf(msg->linseg[i].ay),
          ._value.az = mvn_ntohf(msg->linseg[i].az)
        };
      }
      bodies->lin = true;
      break;

    case 22: /* Angular segment kinematics */
      for(i = 0; i < msg->nitems; i++) {
        id = ntohl(msg->angseg[i].id) - 1;
        if (id >= bodies->segments._length) continue;

        bodies->segments._buffer[id].att = (optional_or_t3d_att){
          ._present = true,
          ._value.qw = mvn_ntohf(msg->angseg[i].qw),
          ._value.qx = mvn_ntohf(msg->angseg[i].qx),
          ._value.qy = mvn_ntohf(msg->angseg[i].qy),
          ._value.qz = mvn_ntohf(msg->angseg[i].qz)
        };
        bodies->segments._buffer[id].avel = (optional_or_t3d_avel){
          ._present = true,
          ._value.wx = mvn_ntohf(msg->angseg[i].wx),
          ._value.wy = mvn_ntohf(msg->angseg[i].wy),
          ._value.wz = mvn_ntohf(msg->angseg[i].wz)
        };
        bodies->segments._buffer[id].aacc = (optional_or_t3d_aacc){
          ._present = true,
          ._value.awx = mvn_ntohf(msg->angseg[i].awx),
          ._value.awy = mvn_ntohf(msg->angseg[i].awy),
          ._value.awz = mvn_ntohf(msg->angseg[i].awz)
        };
      }
      bodies->ang = true;
      break;
  }

  return 0;
}
