/*
 * Copyright (c) 2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Sep  8 2021
 */

#include "acxsensmvn.h"

#include <err.h>
#include <errno.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>

#include "xsensmvn_c_types.h"
#include "codels.h"


/* --- Task recv -------------------------------------------------------- */


/** Codel mvn_recv_start of task recv.
 *
 * Triggered by xsensmvn_start.
 * Yields to xsensmvn_init.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_recv_start(xsensmvn_ids *ids, const xsensmvn_segments *segments,
               const genom_context self)
{
  const char **n;
  size_t i;

  *ids = (xsensmvn_ids){
    .remote = NULL,
    .stream = NULL,
    .localhost = {
      .host = "",	/* automatic guess from getsockname(2) on remote */
      .port = "9763",	/* MVN default */
    }
  };

  if (genom_sequence_reserve(&ids->bodies.segments, mvnsegnames)) {
    warn("malloc");
    return mvn_e_sys_error(ENOMEM, "malloc", self);
  }
  ids->bodies.segments._length = ids->bodies.segments._maximum;
  for(i = 0; i < ids->bodies.segments._length; i++)
    ids->bodies.segments._buffer[i] = (xsensmvn_segment_s){
      .pos._present = false,
      .att._present = false,
      .vel._present = false,
      .avel._present = false,
      .acc._present = false,
      .aacc._present = false
    };

  /* create ports */
  for(i = 0, n = mvnsegname; i < mvnsegnames; i++, n++) {
    if (!*n) continue;
    if (segments->open(*n, self)) {
      errno = EIO;
      warn("%s", *n);
      return mvn_e_sys_error(EIO, *n, self);
    }
  }

  return xsensmvn_init;
}


/** Codel mvn_recv_init of task recv.
 *
 * Triggered by xsensmvn_init.
 * Yields to xsensmvn_poll.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_recv_init(xsensmvn_stream_s **stream,
              const xsensmvn_localhost_s *localhost,
              const genom_context self)
{
  xsensmvn_e_sys_detail d;
  xsensmvn_stream_s *s;

  if (*stream)
    mvn_recv_stop(stream, localhost, self);

  s = malloc(sizeof(*s));
  if (!s) return mvn_e_sys_error(errno, "malloc", self);

  s->fd = mvn_stream_socket(localhost->port, d.what, sizeof(d.what));
  if (s->fd < 0) {
    warnx("%s", d.what);
    free(s);
    return xsensmvn_poll;
  }

  *stream = s;
  warnx("receiving on port %s", localhost->port);
  return xsensmvn_poll;
}


/** Codel mvn_recv_poll of task recv.
 *
 * Triggered by xsensmvn_poll.
 * Yields to xsensmvn_poll, xsensmvn_recv, xsensmvn_init.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_recv_poll(const xsensmvn_stream_s *stream,
              const genom_context self)
{
  struct pollfd fds = {
    .fd = stream ? stream->fd : -1,
    .events = POLLIN
  };
  (void)self;

  if (poll(&fds, 1, 500/*ms*/) == 1)
    return xsensmvn_recv;

  return stream ? xsensmvn_poll : xsensmvn_init;
}


/** Codel mvn_recv of task recv.
 *
 * Triggered by xsensmvn_recv.
 * Yields to xsensmvn_publish, xsensmvn_poll, xsensmvn_init.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_recv(const xsensmvn_stream_s *stream, xsensmvn_body_s *bodies,
         const genom_context self)
{
  char data[2048];
  ssize_t s;
  (void)self;

  if (!stream) return xsensmvn_poll;

  s = recv(stream->fd, data, sizeof(data), 0);
  if (s > 0)
    mvn_stream_decode(data, s, bodies);

  if (bodies->lin && bodies->ang) return xsensmvn_publish;
  return xsensmvn_poll;
}


/** Codel mvn_publish of task recv.
 *
 * Triggered by xsensmvn_publish.
 * Yields to xsensmvn_poll.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_publish(xsensmvn_body_s *bodies, const xsensmvn_segments *segments,
            const genom_context self)
{
  or_pose_estimator_state *state;
  size_t i;

  for(i = 0; i < bodies->segments._length; i++) {
    if (!mvnsegname[i]) continue;

    state = segments->data(mvnsegname[i], self);
    if (!state) continue;

    state->ts = bodies->ts;
    state->intrinsic = false;
    state->pos = bodies->segments._buffer[i].pos;
    state->att = bodies->segments._buffer[i].att;
    state->vel = bodies->segments._buffer[i].vel;
    state->avel = bodies->segments._buffer[i].avel;
    state->acc = bodies->segments._buffer[i].acc;
    state->aacc = bodies->segments._buffer[i].aacc;

    segments->write(mvnsegname[i], self);
  }

  bodies->lin = bodies->ang = false;
  return xsensmvn_poll;
}


/** Codel mvn_recv_stop of task recv.
 *
 * Triggered by xsensmvn_stop.
 * Yields to xsensmvn_ether.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_recv_stop(xsensmvn_stream_s **stream,
              const xsensmvn_localhost_s *localhost,
              const genom_context self)
{
  (void)self;

  if (*stream) {
    close((*stream)->fd);
    free(*stream);
    *stream = NULL;
    warnx("stopped receiving on port %s", localhost->port);
  }

  return xsensmvn_ether;
}


/* --- Activity set_localhost ------------------------------------------- */

/** Codel mvn_set_localhost of activity set_localhost.
 *
 * Triggered by xsensmvn_start.
 * Yields to xsensmvn_ether.
 * Throws xsensmvn_e_sys.
 */
genom_event
mvn_set_localhost(xsensmvn_stream_s **stream,
                  xsensmvn_localhost_s *newhost,
                  xsensmvn_localhost_s *localhost,
                  const genom_context self)
{
  xsensmvn_e_sys_detail d;

  if (mvn_stream_gethost(newhost->host, sizeof(newhost->host),
                         newhost->port, sizeof(newhost->port),
                         d.what, sizeof(d.what))) {
    d.code = EIO;
    return xsensmvn_e_sys(&d, self);
  }

  mvn_recv_stop(stream, localhost, self);
  *localhost = *newhost;
  return xsensmvn_ether;
}
